"""Mass Memory subsystem"""
import array
import pathlib
import time
from rsfsup.common import Subsystem, validate


class MassMemory(Subsystem, kind="Mass Memory"):
    """Mass memory subsystem

    Attributes:
        instr (Fsup)
    """

    @staticmethod
    def process_catalog(cat):
        """Process comma separated q-string files into a list

        Parameters:
            cat (str): comma separated q-string files
        """
        cat = cat.replace("'", "")
        return cat.split(",")

    @property
    def drive(self):
        """value (str): {D, F}"""
        res = self._visa.query("MMEMORY:MSIS?")
        return res.replace("'", "")

    @drive.setter
    @validate
    def drive(self, value):
        self._visa.write(f"MMEMORY:MSIS '{value}:'")

    @property
    def directory(self):
        """value (str): directory, e.g. 'D:\\USER'"""
        res = self._visa.query("MMEMORY:CDIRECTORY?")
        return res.replace("'", "")

    @directory.setter
    @validate
    def directory(self, value):
        self._visa.write(f"MMEMORY:CDIRECTORY '{value}'")

    @property
    def file_listing(self):
        """(list): list of files in current directory"""
        cat = self._visa.query("MMEMORY:CATALOG?")
        return MassMemory.process_catalog(cat)

    def get(self, source, destination=None):
        """Get the source file and save at destination

        Transfers the source file raw bytes to the control computer as unsigned
        8-bit integers.

        Parameters:
            source (str): file to retrieve from the instrument
            destination (str): path and file name to save on control computer
        """
        qstr = f"MMEMORY:DATA? '{source}'"
        raw_bytes = self._visa.query_binary_values(qstr, datatype="B")
        if destination is None:
            return raw_bytes
        with open(destination, "wb") as f:
            raw_array = array.array("B")
            raw_array.fromlist(raw_bytes)
            f.write(raw_array)
        return None

    def put(self, source, destination):
        """Put the source file on the instrument at destination

        Parameters:
            source (str): file on control computer
            destination (str): path and file name to save on the instrument
        """
        with open(source, "rb") as f:
            file_bytes = f.read()
        destination = pathlib.PurePath(destination)
        if destination.is_absolute():
            self.drive = destination.drive
            self.directory = str(destination.parent)
        else:
            self.drive = "D:"
        prefix = f"MMEMORY:DATA '{destination}',#{len(str(len(file_bytes)))}{len(file_bytes)}"
        prefix = prefix.encode("utf-8")
        suffix = "; *OPC".encode("utf-8")
        self._visa.write("*CLS")
        self._visa.write_raw(prefix + file_bytes + suffix)
        while not self._visa.stb & 32:
            time.sleep(1)
        if self._visa.stb & 4:
            err = self._visa.query("SYSTEM:ERROR?")
            raise ValueError(err)

    def delete(self, file):
        """Delete file

        Parameters:
            file (str): file to delete
        """
        self._visa.write("*CLS")
        self._visa.write(f"MMEMORY:DELETE '{file}'; *OPC")
        while not self._visa.stb & 32:
            time.sleep(1)
        if self._visa.stb & 4:
            err = self._visa.query("SYSTEM:ERROR?")
            raise ValueError(err)



"""Bandwidth subsystem"""
from rsfsup.common import Subsystem, scale_frequency, validate


class Bandwidth(Subsystem, kind="BW"):
    """Bandwidth subsystem

    Attributes:
        instr (Fsup)
    """

    def __init__(self, instr):
        super().__init__(instr)
        # For now, adjust resolution and video bandwidths automatically according to
        # the span.
        self._visa.write("BANDWIDTH:RESOLUTION:AUTO ON")
        self._visa.write("BANDWIDTH:RESOLUTION:TYPE NORMAL")
        self._visa.write("BANDWIDTH:VIDEO:AUTO ON")

    @property
    def resolution_bandwidth(self):
        """Resolution bandwidth

        Parameters
        ----------
        value : str {AUTO, 10 Hz to 50 MHz}
            When set to AUTO, RBW is based on the RBW/SPAN ratio.
        """
        value = self._visa.query(f"SENSE{self._screen()}:BANDWIDTH:RESOLUTION?")
        return scale_frequency(float(value))

    @resolution_bandwidth.setter
    @validate
    def resolution_bandwidth(self, value):
        if value == "AUTO":
            self._visa.write(f"SENSE{self._screen()}:BAND:RES:AUTO ON")
        else:
            self._visa.write(f"SENSE{self._screen()}:BAND:RES:AUTO OFF")
            self._visa.write(f"SENSE{self._screen()}:BAND:RES {value}")

    @property
    def video_bandwidth(self):
        """Video bandwidth

        Parameters
        ----------
        value : str {AUTO, 1 Hz to 10 MHz
            When set to AUTO, video bandwidth is based on the VideoBandwidth/RBW ratio.
        """
        value = self._visa.query(f"SENSE{self._screen()}:BANDWIDTH:VIDEO?")
        return scale_frequency(float(value))

    @video_bandwidth.setter
    @validate
    def video_bandwidth(self, value):
        if value == "AUTO":
            self._visa.write(f"SENSE{self._screen()}:BAND:VID:AUTO ON")
        else:
            self._visa.write(f"SENSE{self._screen()}:BAND:VID:AUTO OFF")
            self._visa.write(f"SENSE{self._screen()}:BAND:VID {value}")



"""Trigger subsystem for spectrum analyzer mode"""
from rsfsup.common import Subsystem, validate


class Trigger(Subsystem, kind="Trigger"):
    """Trigger subsystem for spectrum analyzer mode

    Specifically supports screen A immediate and external triggering.

    Attributes:
        instr (Fsup)
    """

    _sources = {"IMM": "FREE RUN", "EXT": "EXTERNAL"}
    _slopes = {"POS": "POSITIVE", "NEG": "NEGATIVE"}

    @property
    def source(self):
        """value (str): {IMMEDIATE, EXTERNAL}"""
        src = self._visa.query("TRIGGER1:SEQUENCE:SOURCE?")
        return Trigger._sources.get(src, src)

    @source.setter
    @validate
    def source(self, value):
        sources = {v: k for k, v in Trigger._sources.items()}
        src = sources.get(value, value)
        self._visa.write(f"TRIGGER1:SEQUENCE:SOURCE {src}")

    @property
    def level(self):
        """value (float, str): external trigger level in volts"""
        res = self._visa.query("TRIGGER1:SEQUENCE:LEVEL:EXTERNAL?")
        return f"{res} V"

    @level.setter
    @validate
    def level(self, value):
        self._visa.write(f"TRIGGER1:SEQUENCE:LEVEL:EXTERNAL {value}")

    @property
    def slope(self):
        """value (str): {POSITIVE, NEGATIVE}"""
        res = self._visa.query("TRIGGER1:SEQUENCE:SLOPE?")
        return Trigger._slopes[res]

    @slope.setter
    @validate
    def slope(self, value):
        self._visa.write(f"TRIGGER1:SEQUENCE:SLOPE {value}")

    @property
    def offset(self):
        """value (float, str): -100 s to 100 s, negative means pretrigger for zero span"""
        res = self._visa.query("TRIGGER1:SEQUENCE:HOLDOFF?")
        return f"{res} s"

    @offset.setter
    @validate
    def offset(self, value):
        self._visa.write(f"TRIGGER1:SEQUENCE:HOLDOFF {value}")
